# DEBUG.pmc
#
# Copyright 2008-2010, Larry Wall
#
# You may copy this software under the terms of the Artistic License,
#     version 2.0 or later.

package DEBUG;
our @EXPORTS = qw(
        autolexer
        symtab
        fixed_length
        fates
        longest_token_pattern_generation
        EXPR
        matchers
        trace_call
        cursors
        try_processing
        mixins
        callm_show_subnames
        use_color
);
use constant {
    autolexer => 1,
    symtab => 2,
    fixed_length => 4,
    fates => 8,
    longest_token_pattern_generation => 16,
    EXPR => 32,
    matchers => 64,
    trace_call=> 128,
    cursors => 256,
    try_processing => 1024,
    mixins => 2048,
    callm_show_subnames => 16384,
    use_color => 32768
};
